### EM algorithm for factor model of Gaussian data
#### Function is adapted from the one in the R cate package, Wang and Zhao (2018)
#### Reference: https://cran.r-project.org/web/packages/cate/
#### Include tolerance parameter for gradient: gtol

fa.em <- compiler::cmpfun(function(Y, r, tol = 1e-6, gtol = 1e-3, maxiter = 5000) {
  
  ## The EM algorithm:
  ##
  ## Y = Z Gamma' + E Sigma^{1/2}                        (n * p)
  ## mle to estimate Gamma (p * r) and Sigma             (p * p)
  
  ## E step:
  ## EZ = Y (Gamma Gamma' + Sigma)^{-1} Gamma            (n * r)
  ## VarZ = I - Gamma' (Gamma Gamma' + Sigma)^{-1} Gamma (r * r)
  ## EZ'Z = n VarZ + EZ' * EZ                            (r * r)
  
  ## M step:
  ## update Gamma: Gamma = (Y' EZ)(EZ'Z)^{-1}
  ## update Sigma:
  ## Sigma = 1/n diag(Y'Y - Gamma EZ' Y - Y' EZ' Gamma' + Gamma EZ'Z Gamma')
  
  ## The log-likelihood (simplified)
  ## llh <- -log det(Gamma Gamma' + Sigma) - tr((Gamma Gamma' + Sigma)^{-1} S)
  ##
  ## (llh + sum(log(sample.var)))*n/2

  ## For details, see http://cs229.stanford.edu/notes/cs229-notes9.pdf
  
  p <- ncol(Y)
  n <- nrow(Y)
  Y = scale(Y,scale = F)
  
  ## precompute quantitites
  I <- diag(rep(1, r))
  ## diagonal of the sample Covariance S
  sample.var <- colMeans(Y^2)
  
  ## initialize parameters
  init <- fa.pc(Y, r)
  Gamma <- sqrt(sample.var)*init$Gamma
  invSigma <- 1/(init$Sigma * sample.var)
  llh <- -Inf # log-likelihood
  
  ## compute quantities needed
  tilde.Gamma <- sqrt(invSigma) * Gamma
  M <- diag(r) + t(tilde.Gamma) %*% tilde.Gamma
  eigenM <- eigen(M, T)
  YSG <- Y %*% (invSigma * Gamma)
  
  ## compute likelihood
  R <- chol(M)
  logdet <- -sum(log(invSigma)) + 2*sum(log(diag(R)))
  llh <- {-logdet}*0.5*n - n*p/2
  
  converged <- FALSE
  for (iter in 1:maxiter) {
    
    ## E step:
    ## Using Woodbury matrix identity:
    ## tilde.Gamma = Sigma^{-1/2} Gamma
    ## VarZ = (I + tilde.Gamma' tilde.Gamma)^{-1}
    ## EZ = Y Sigma^{-1} Gamma VarZ
    varZ <- eigenM$vectors %*% (1/eigenM$values * t(eigenM$vectors))
    EZ <- YSG %*% varZ
    EZZ <- n * varZ + t(EZ) %*% EZ
    
    ## M step:
    eigenEZZ <- eigen(EZZ, symmetric = TRUE)
    YEZ <- t(Y) %*% EZ
    ## EZ'Z = G'G
    G <- sqrt(eigenEZZ$values) * t(eigenEZZ$vectors)
    ## updating invSigma
    old.invSigma <- invSigma
    invSigma <- 1/(sample.var - 2/n * rowSums(YEZ * Gamma) +
                     1/n * rowSums((Gamma %*% t(G))^2))
    ## updating Gamma
    Gamma <- YEZ %*% eigenEZZ$vectors %*%
      (1/eigenEZZ$values * t(eigenEZZ$vectors))
    
    ## compute quantities needed
    tilde.Gamma <- sqrt(invSigma) * Gamma
    M <- diag(r) + t(tilde.Gamma) %*% tilde.Gamma
    eigenM <- eigen(M, T)
    YSG <- Y %*% (invSigma * Gamma)
    
    ## compute likelihood and check for convergence
    old.llh <- llh
    R <- chol(M)
    logdet <- -sum(log(invSigma)) + 2*sum(log(diag(R)))
    llh <- {-logdet}*0.5*n - n*p/2
    
    g <- (rowSums(Gamma^2) + 1/invSigma)/sample.var - 1
    
    if (abs(llh - old.llh) < tol * abs(llh) && sum(abs(g)) < gtol) {
      converged <- TRUE
      break
    }
  }
  
  em <- list(Gamma = Gamma,
             Sigma = 1/invSigma,
             Var = sample.var,
             #loglik = (llh + sum(log(sample.var)))*n/2,
             loglik = llh,
             niter = iter,
             converged = converged)
  em$BIC = -2*em$loglik + r*p*log(n);
  return(em)
  
})