### 100 Simulation runs for data-driven case with:
#### Fitted parameters for the suicide attempter data, Just, et al. (2017)
#### Reference for data application: Dai, Dutta and Maitra (2019).

#### Required packages:
##### fad: See install instructions in install.R
##### doMC: Foreach parallel adaptor for 'parallel'

## Load packages and source codes
library("fad") 
library("doMC")
registerDoMC(16) # Parelled with 16 cores.

source("fapc.R") # Initialization for EM algorithm
source("faem.R") # EM algorithm
source("postdiag.R") # Matrix-vector multiplication X %*% v, data generation step

## Load fitted parameters
### L and D are scaled
load("data-L.rda")
load("data-D.rda")
load("data-SD.rda")

Sim <- 100 # Simulation runs
N <- 180 # Sample size for suicide attempters
P <- 24547 # Data dimension
Q <- 2 # Data for suicide attempters are fitted with 2 factors
Mq <- 2*Q # Maximum number of factors for simulation

## Save multiple outputs of foreach dopar loop
multiResultClass <- function(SD=NULL,L=NULL,D=NULL,X=NULL,
                             r1=NULL,t1=NULL,r2=NULL,t2=NULL)
{
  m <- list(SD = SD, L = L, D = D, X = X, 
            r1 = r1, t1 = t1,r2 = r2, t2 = t2)
  class(m) <- append(class(m),"multiResultClass")
  return(m)
}

RES <- foreach(i=1:Sim) %dopar% {
  result <- multiResultClass()
  set.seed(10+i) # For reproduced results
  
  ## Generate data with:
  ### Fitted correlation parameters L: factor loadings, D: uniqueness
  ### Fitted standard deviations: SD
  ### Random factors follows multivariate normal with 0 means and identity covariance matrix
  ### Note: Means are ignored since the MLE correspond to the sample means
  L <- suicide.loadings$a
  D <- suicide.D$a
  SD <- suicide.SD$a
  
  # Obtain covariance parameters
  L <- SD*L
  D <- SD^2*D
  X <- matrix(rnorm(N*Q),N,Q)
  X <- tcrossprod(X,L) + postmdiag(matrix(rnorm(N*P),N,P), sqrt(D))
  
  t1 = t2 <- rep(0,Mq)
  r1 = r2 <- list()
  for(q in 1:Mq){
    ## FAD
    ### For stopping criteria used in Dai, Dutta and Maitra (2019):
    ### factr = 100 in optim_share function (default in fad) 
    ptm <- proc.time()
    res1 <- fad(x = X, factors = q)
    time <- proc.time() - ptm
    t1[q] <- time[3]
    r1[[q]] <- res1

    ## EM
    ### For comparable stopping criteria with FAD:
    ### Control the tolerance parameter for gradient 
    ### Note: Maximum number of iterations = 5000
    ptm <- proc.time()
    res2 <- fa.em(Y = X, r = q, gtol = res1$gerr)
    time <- proc.time() - ptm
    t2[q] <- time[3]
    r2[[q]] <- res2
  }
  
  ## Collect results for each loop
  result$SD <- SD
  result$L <- L
  result$D <- D
  result$X <- X
  result$r1 <- r1
  result$t1 <- t1
  result$r2 <- r2
  result$t2 <- t2

  return(result)
}

save(RES, file = "simp-da-res.rda")
