### 100 Simulation runs for a high noise model with:
#### The uniquenesses follows unit means and variances of 0
#### n = 100, p = 5000, q = 3
#### Reference for simulation setup: Owen and Wang (2015).

#### Required packages:
##### SMFilter, invgamma: For data generation
##### fad: See install instructions in install.R
##### doMC: Foreach parallel adaptor for 'parallel'

## Load packages and source codes
library("SMFilter") 
library("invgamma") 
library("fad") 
library("doMC")
registerDoMC(16) # Parelled with 16 cores.

source("fapc.R") # Initialization function for EM algorithm
source("faem.R") # EM algorithm
source("postdiag.R") # Matrix-vector multiplication X %*% v, data generation step

Sim <- 100 # Simulation runs
N <- 100 # Sample size
P <- 5000 # Data dimension
Q <- 3 # True number of factors
Mq <- 2*Q # Maximum number of factors fitted 

## Save multiple outputs of foreach dopar loop
multiResultClass <- function(X=NULL,Xs=NULL,D=NULL,L=NULL,
                             r1=NULL,t1=NULL,r2=NULL,t2=NULL)
{
  m <- list(X = X, Xs = Xs,D = D, L = L,
            r1 = r1, t1 = t1,r2 = r2, t2 = t2)
  class(m) <- append(class(m),"multiResultClass")
  return(m)
}

RES <- foreach(i=1:Sim) %dopar% {
  result <- multiResultClass()
  set.seed(10+i) # For reproduced results
  
  ## Generate data with covariance parameters
  ### L: factor loadings, D: uniqueness
  ### Follow the simulation setup in Owen and Wang (2015).
  D <- rep(1,P) # unit means and variances of 0
  E <- postmdiag(matrix(rnorm(N*P),N,P), sqrt(D))
  ## Draws a sample from the uniform distribution on the Stiefel manifold.
  V <- matrix(SMFilter::runif_sm(1,ip = N, ir = Q),N,Q) 
  U <- matrix(SMFilter::runif_sm(1,ip = P, ir = Q),P,Q)
  gamma <- P/N
  d <- sqrt(runif(Q,0.1,sqrt(gamma)/2))
  L <- (N^0.25)*postmdiag(sqrt(D)*U,d)  
  Xs <- sqrt(N)*V%*%(d*t(sqrt(D)*U)) # Decouple signal and noise.
  X <- Xs + E
  
  t1 = t2 <- rep(0,Mq)
  r1 = r2 <- list()
  for(q in 1:Mq){
    ## FAD
    ### For stopping criteria used in Dai, Dutta and Maitra (2019):
    ### factr = 100 in optim_share function (default in fad) 
    ptm <- proc.time()
    res1 <- fad(x = X, factors = q)
    time <- proc.time() - ptm
    t1[q] <- time[3]
    r1[[q]] <- res1
    
    ## EM
    ### For comparable stopping criteria with FAD:
    ### Control the tolerance parameter for gradient 
    ### Note: Maximum iterations = 5000
    ptm <- proc.time()
    res2 <- fa.em(Y = X, r = q, gtol = res1$gerr)
    time <- proc.time() - ptm
    t2[q] <- time[3]
    r2[[q]] <- res2
  }
  
  ## Collect results for each loop
  result$L <- L
  result$D <- D
  result$X <- X
  result$Xs <- Xs
  result$r1 <- r1
  result$t1 <- t1
  result$r2 <- r2
  result$t2 <- t2

  return(result)
}

## Save results
save(RES, file = "simp-hnoise1v0-res.rda")
